/*
  CatWeasel MK3 PCI Linux driver (c) 2002 by Dirk Jagdmann
  <doj@cubic.org>

  This file is part of CatWeasel MK3 PCI Linux driver.

  The CatWeasel MK3 PCI Linux driver is free software; you can
  redistribute it and/or modify it under the terms of the GNU General
  Public License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  The CatWeasel MK3 PCI Linux driver is distributed in the hope that
  it will be useful, but WITHOUT ANY WARRANTY; without even the
  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
  PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with the CatWeasel MK3 PCI Linux driver; if not, write to the
  Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
  MA 02111-1307 USA

  $Header: /home/cvs/sidblaster/device/cwsid.c,v 1.7 2003/09/10 17:01:35 doj Exp $
*/

/*
  Code interfacing the SID
*/

#include <cw.h>
#include <cwsid.h>

#include <asm/io.h>
#include <asm/uaccess.h>
#include <linux/delay.h>
#include <linux/miscdevice.h>
#include <linux/poll.h>

static int cwsid_irq=-1;
static int cwsid_ntsc=0;
static int cwsid_hz=-1;

static int cwsid_delay = 2;
MODULE_PARM(cwsid_delay,"i");
MODULE_PARM_DESC(cwsid_delay,"delay in usec after SID accesses (default=2)");

DECLARE_WAIT_QUEUE_HEAD (sid_queue);

static unsigned char sid_buffer[CWSID_MAX_NUM*0x20];
/*static void *sid_id;		used by IRQs */


static void sid_write(unsigned char addr, unsigned char data, int sid)
{
  if(addr<0x19 && sid<cw_num)
    {
      addr|=cwsid_ntsc;
      outb(data, cw_register[sid][SidDat]);
      outb(addr, cw_register[sid][SidCommand]);
      udelay(cwsid_delay);
    }
}

unsigned char sid_read(unsigned char addr, int sid)
{
  if(addr>=0x19 && addr<=0x1C && sid<cw_num)
    {
      addr|=cwsid_ntsc|0x20;
      outb(addr, cw_register[sid][SidCommand]);
      udelay(cwsid_delay);
      return inb(cw_register[sid][SidDat]);
    }
  return 0;
}

static void sid_reset()
{
  int i, s;
  cwsid_ntsc=0;
  for(s=0; s<cw_num; s++)
    for(i=0; i<0x19; i++)
      sid_write(i, 0, s);
  memset(sid_buffer, 0, sizeof(sid_buffer));
}

static int cwsid_write(struct file *file, const char *buf, size_t size, loff_t *pos)
{
  int max, i;

  if (verify_area(VERIFY_READ, buf, size)) 
    return -EINVAL;

  max=(file->f_pos+size>sizeof(sid_buffer))?(sizeof(sid_buffer)-file->f_pos):size;
  if(max==1)
    get_user(sid_buffer[file->f_pos], buf);
  else
    copy_from_user(sid_buffer + file->f_pos, buf, max);

  for(i=file->f_pos; i<file->f_pos+max; i++)
    sid_write(i&0x1F, sid_buffer[i], i>>5);

  file->f_pos += max;
  return max;
}

static int cwsid_read(struct file *file, char *buf, size_t size, loff_t *pos)
{
  int max, i;

  if (verify_area(VERIFY_WRITE, buf, size)) 
    return -EINVAL;

  max=(file->f_pos+size>sizeof(sid_buffer))?(sizeof(sid_buffer)-file->f_pos):size;

  for(i=file->f_pos; i<file->f_pos+max; i++)
    if((i&0x1F)>=0x19 && (i&0x1F)<=0x1C)
      sid_buffer[i]=sid_read(i&0x1F, i>>5);
  copy_to_user(buf, sid_buffer+file->f_pos, max);
  
  file->f_pos += max;
  return max;
}

static int cwsid_open(struct inode *inode, struct file *file)
{
  MOD_INC_USE_COUNT;
  return 0;
}

static int cwsid_close(struct inode *inode, struct file *file)
{
  MOD_DEC_USE_COUNT;
  if(!MOD_IN_USE)
    sid_reset();
  return 0;
}

static int cwsid_ioctl(struct inode *node, struct file *file, unsigned int cmd, unsigned long arg)
{
  switch(cmd)
    {
    case CWSID_IOCTL_RESET: 
      sid_reset(); 
      return 0;
	  
    case CWSID_IOCTL_SIDTYPE:
    case CWSID_IOCTL_CARDTYPE:
      put_user(CWSID_MAGIC, (int*)arg);
      return 0;

    case CWSID_IOCTL_IRQAVAIL:
      put_user((cwsid_irq>=0)?1:0, (int*)arg);
      return 0;

    case CWSID_IOCTL_WAITIRQ:
      if(cwsid_irq>=0)
	{
	  interruptible_sleep_on(&sid_queue); 
	  return 0;
	}
      break;
      
    case CWSID_IOCTL_PAL:
      cwsid_ntsc=0;
      return 0;
      
    case CWSID_IOCTL_NTSC:
      cwsid_ntsc=0x40;
      return 0;
      
    case CWSID_IOCTL_SETHZ:
      if(cwsid_irq>=0)
	{
	  if(arg!=cwsid_hz)
	    {
	      cwsid_hz=arg;
	      /* reinit irq machine */
	    }
	  return 0;
	}
      return -1;
      
    case CWSID_IOCTL_DELAY:
      cwsid_delay=arg;
      return 0;

    }
  
  return -EINVAL;
}

static unsigned int cwsid_select(struct file *file, poll_table *wait)
{
  if(cwsid_irq>=0)
    {
      interruptible_sleep_on(&sid_queue); 
      return 1;
    }
  return 0;
}

static struct file_operations cwsid_fops = {
  owner: THIS_MODULE,
  read: cwsid_read,
  write: cwsid_write,
  poll: cwsid_select,
  ioctl: cwsid_ioctl,
  open: cwsid_open,
  release: cwsid_close,
  /*llseek: cwsid_lseek,*/
};

static struct miscdevice cwsid_miscdevice = {
  minor: CWSIDMINOR,
  name: "sid",
  fops: &cwsid_fops,
};

int __init cwsid_init()
{
  if(misc_register(&cwsid_miscdevice) < 0)
    {
      printk(KERN_ERR DEVICE_NAME ": register failed (already in use ?)\n");
      return -EIO;
    }

  sid_reset();
  return 0;
}

void __exit cwsid_exit()
{
  if(misc_deregister(&cwsid_miscdevice) < 0)
    printk(KERN_ERR DEVICE_NAME ": unregister failed\n");
  
  sid_reset();
}
