/*
  sid (c) 1999-2003 by Dirk Jagdmann <doj@cubic.org>
  The 6510 emulation code is (c) 1999 by Dirk Jadgmann and Tammo Hinrichs

  This file is part of sid.

  sid is free software; you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation; either version 2 of the License, or (at your
  option) any later version.

  sid is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
  License for more details.

  You should have received a copy of the GNU General Public License
  along with sid; if not, write to the Free Software Foundation, Inc.,
  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
  
  $Header: /home/cvs/sidblaster/sid/6510.h,v 1.9 2003/05/21 23:10:32 doj Exp $
*/

#ifndef MOS6510__H
#define MOS6510__H

#include <stdio.h>
#include <set>

#include "defines.h"

class C64;

// positions of the flags in the flag register
enum flag_t {
  FLAG_C=1,
  FLAG_Z=2,
  FLAG_I=4,
  FLAG_D=8,
  FLAG_B=16,
  FLAG_V=64,
  FLAG_N=128
};

enum opcode_t {
  ADC, AND, ASL, BCC, BCS, BEQ, BIT, BMI, BNE, BPL, BRK, BVC, BVS, CLC,
  CLD, CLI, CLV, CMP, CPX, CPY, DEC, DEX, DEY, EOR, INC, INX, INY, JMP,
  JSR, LDA, LDX, LDY, LSR, NOP, ORA, PHA, PHP, PLA, PLP, ROL, ROR, RTI,
  RTS, SBC, SEC, SED, SEI, STA, STX, STY, TAX, TAY, TSX, TXA, TXS, TYA,
  ASO, RLA, LSE, RRA, AXS, LAX, DCM, INS,
  ALR, ARR, XAA, OAL, SAX, SKB, SKW, HLT,
  TAS, SAY, XAS, AXA, ANC, LAS
};

enum addrmode_t {
  IMP, IMM, ABS, ABSX, ABSY, ZP, ZPX, ZPY, IND, INDX, INDY, ACC, REL
};



/// a MOS6510 CPU
class MOS6510
{
 private:
  MOS6510();

 protected:
  /// the computer object into which this CPU is built
  C64 *comp;

  /// the amount of cycles in the current function call
  int cycles;

  /// the amount of cycles of the current opcode
  int cyclesSingle;

  /// the amount of cycles since the last reset
  int cyclesTotal;

  /// the amount of instructions in the current function call
  int instructions;

  /// the amount of instructions since the last reset
  int instructionsTotal;
  
  /// register
  byte a,x,y,s,p;
  /// register 
  word pc;

  /// container type for breakpoints
  typedef std::set<word> breakpointlist_t;
  /// container storing breakpoints
  breakpointlist_t breakpointlist;

  /** fetches the content of the adress used in the currently processed
      instruction. the amount of cycles used is counted
      @param mode defines the adress-mode used by the processor
      @return a byte containing the content of the indexed adress. if the adessing mode was illegal 0 is returned
  */
  const byte getaddr(const int mode);

  /** writes val into the location pointed to by the currently
      processed instruction. the adressing mode is determined by
      mode. this funtion is used by opcodes that work on memory
      directly (inc, dec), because the program counter (pc) is not
      incremented.<br>

      See also void putaddr(int mode, byte val)
      @param mode address mode
      @param val byte to write into memory
  */
  void setaddr(const int mode, const byte val);

  /** writes val into the memory location pointed to by the operandes
      of the currently processed instruction. the adressing mode is
      determined by mode.<br>

      See also "static void setaddr(int mode, byte val)"

      @param mode address mode
      @param val byte to write into memory
  */
  void putaddr(const int mode, const byte val);

  /** sets flag to either true or false depending on cond.
      you should use the flag definitions like FLAG_I or otherwise the
      flag register may be corrupted after this function call.

      @param flag flag to process
      @param cond set or reset flag
  */
  void setflags(const int flag, const bool cond)
    {
      cond?p|=flag:p&=~flag;
    }

  /** pushes the val onto the stack
      @param val a byte to push
  */
  void push(const byte val);

  /** pops a byte from the stack
      @return a byte popped from the stack
  */
  const byte pop();

  /** performs a branch if flag is true. the branch address if fetched
      via "immediate adressing mode"

      @param flag condition onto to branch
  */
  void branch(const bool flag);

  /** prints the contents of the cpu registers
      @param f FILE stream used to output
  */
  void status(FILE *f=stdout) const;


 public:

  /// true for debug mode allowing single stepping of code and extended abortion.
  bool debug;
  /// abort execution immediatly
  volatile bool abort;
  
  /** disassembles the current instruction pointed by loc
      @param loc memory location or -1 for pc
      @param f FILE stream used to output
      @return the offset to the next opcode
  */
  const int disass(int loc=-1, FILE *f=stdout) const;

  /** makes a reset and lets you load the cpu registers with predefined values. If npc is not given the pc is loaded with the word stored at 0xfffc in memory.
      @param npc the content of the program counter (pc)
      @param na the content of the accumulator (a)
      @param nx the content of the x-register (x)
      @param ny the content of the y-register (y)
  */
  void reset(const int npc=-1, const byte na=0, const byte nx=0, const byte ny=0);

  /** dumps a status of the cpu. the contents of the registers and the
      currently processed opcode is shown.

      @param f FILE stream used to output
  */
  void info(FILE *f=stdout) const;

  /** executes one opcode
      @return number of cycles used to process the current opcode. -1 on fatal error
  */
  const int parse();

  /** lets you execute a subroutine starting in memory at npc.
      jsr performs a cpu reset and pushes two 0 onto the stack. It then jumps
      into the routine starting at npc. Because the stack contains two zeros
      the subroutine can exit either via a RTS instruction.

      @param npc the content of the program counter (pc), -2 to load from reset vector 0xFFFC, -1 to for current pc
      @param na the content of the accumulator (a), -1 for current a
      @param nx the content of the x-register (x), -1 for current x
      @param ny the content of the y-register (y), -1 for current y
      @param maxCycles maximum number of cycles the CPU should run. -1 for unlimited.
      @return the number of cycles used by the executed routine
  */
  const int jsr(const int npc=-2, const int na=0, const int nx=0, const int ny=0, const int maxCycles=-1);

  /// @return the number of cycles of the last jsr()
  const int CPUcycles() const { return cycles; }

  /// @return the number of cycles since last reset
  const int CPUcyclesTotal() const { return cyclesTotal; }

  /// @return number of instructions of the last jsr()
  const int CPUinstructions() const { return instructions; }

  /// @return number of instructions since last reset
  const int CPUinstructionsTotal() const { return instructionsTotal; }


  MOS6510(C64 *c) : 
    comp(c),
    debug(false),
    abort(false)
    {
      reset();
    }

  /** set registers */
  void set(const word npc, const byte na, const byte nx, const byte ny)
    {
      pc=npc; a=na; x=nx; y=ny;
    }

  /**
     toggle breakpoint at addr
     @param addr address of breakpoint, -1 for pc
  */
  void breakpointToggle(int addr);

  /// clear all breakpoints
  void breakpointClear() { breakpointlist.clear(); }

  /// print all breakpoints to stdout
  void breakpointPrint();
 protected:
  /// @return true if a breakpoint exists at addr
  bool breakpoint(int addr);
 public:

};

#endif
